@@ -1,17 +1,19 @@
-use super::{stars, DifficultyAttributes};
-use crate::{Beatmap, Mods, PpResult, StarResult};
+use super::{stars, FruitsDifficultyAttributes, FruitsPerformanceAttributes, FruitsScoreState};
+use crate::{Beatmap, DifficultyAttributes, Mods, PerformanceAttributes};
 
-/// Calculator for pp on osu!ctb maps.
+/// Performance calculator on osu!ctb maps.
 ///
 /// # Example
 ///
 /// ```
-/// # use rosu_pp::{FruitsPP, PpResult, Beatmap};
+/// use rosu_pp::{FruitsPP, Beatmap};
+///
 /// # /*
 /// let map: Beatmap = ...
 /// # */
 /// # let map = Beatmap::default();
-/// let pp_result: PpResult = FruitsPP::new(&map)
+///
+/// let pp_result = FruitsPP::new(&map)
 ///     .mods(8 + 64) // HDDT
 ///     .combo(1234)
 ///     .misses(1)
@@ -30,23 +32,24 @@ use crate::{Beatmap, Mods, PpResult, StarResult};
 /// ```
 #[derive(Clone, Debug)]
 #[allow(clippy::upper_case_acronyms)]
-pub struct FruitsPP<'m> {
-    map: &'m Beatmap,
-    attributes: Option<DifficultyAttributes>,
+pub struct FruitsPP<'map> {
+    map: &'map Beatmap,
+    attributes: Option<FruitsDifficultyAttributes>,
     mods: u32,
     combo: Option<usize>,
 
-    n_fruits: Option<usize>,
-    n_droplets: Option<usize>,
-    n_tiny_droplets: Option<usize>,
-    n_tiny_droplet_misses: Option<usize>,
-    n_misses: usize,
+    pub(crate) n_fruits: Option<usize>,
+    pub(crate) n_droplets: Option<usize>,
+    pub(crate) n_tiny_droplets: Option<usize>,
+    pub(crate) n_tiny_droplet_misses: Option<usize>,
+    pub(crate) n_misses: usize,
     passed_objects: Option<usize>,
 }
 
-impl<'m> FruitsPP<'m> {
+impl<'map> FruitsPP<'map> {
+    /// Create a new performance calculator for osu!ctb maps.
     #[inline]
-    pub fn new(map: &'m Beatmap) -> Self {
+    pub fn new(map: &'map Beatmap) -> Self {
         Self {
             map,
             attributes: None,
@@ -62,9 +65,7 @@ impl<'m> FruitsPP<'m> {
         }
     }
 
-    /// [`FruitsAttributeProvider`] is implemented by [`DifficultyAttributes`](crate::fruits::DifficultyAttributes),
-    /// [`StarResult`](crate::StarResult), and by [`PpResult`](crate::PpResult) meaning you can give the
-    /// result of a star calculation or a pp calculation.
+    /// Provide the result of a previous difficulty or performance calculation.
     /// If you already calculated the attributes for the current map-mod combination,
     /// be sure to put them in here so that they don't have to be recalculated.
     #[inline]
@@ -135,6 +136,10 @@ impl<'m> FruitsPP<'m> {
     }
 
     /// Amount of passed objects for partial plays, e.g. a fail.
+    ///
+    /// If you want to calculate the performance after every few objects, instead of
+    /// using [`FruitsPP`] multiple times with different `passed_objects`, you should use
+    /// [`FruitsGradualPerformanceAttributes`](crate::fruits::FruitsGradualPerformanceAttributes).
     #[inline]
     pub fn passed_objects(mut self, passed_objects: usize) -> Self {
         self.passed_objects.replace(passed_objects);
@@ -142,16 +147,34 @@ impl<'m> FruitsPP<'m> {
         self
     }
 
+    /// Provide parameters through an [`FruitsScoreState`].
+    #[inline]
+    pub fn state(mut self, state: FruitsScoreState) -> Self {
+        let FruitsScoreState {
+            max_combo,
+            n_fruits,
+            n_droplets,
+            n_tiny_droplets,
+            n_tiny_droplet_misses,
+            misses,
+        } = state;
+
+        self.combo = Some(max_combo);
+        self.n_fruits = Some(n_fruits);
+        self.n_droplets = Some(n_droplets);
+        self.n_tiny_droplets = Some(n_tiny_droplets);
+        self.n_tiny_droplet_misses = Some(n_tiny_droplet_misses);
+        self.n_misses = misses;
+
+        self
+    }
+
     /// Generate the hit results with respect to the given accuracy between `0` and `100`.
     ///
     /// Be sure to set `misses` beforehand! Also, if available, set `attributes` beforehand.
-    pub fn accuracy(mut self, mut acc: f32) -> Self {
+    pub fn accuracy(mut self, mut acc: f64) -> Self {
         if self.attributes.is_none() {
-            self.attributes.replace(
-                stars(self.map, self.mods, self.passed_objects)
-                    .attributes()
-                    .unwrap(),
-            );
+            self.attributes = Some(stars(self.map, self.mods, self.passed_objects));
         }
 
         let attributes = self.attributes.as_ref().unwrap();
@@ -160,9 +183,10 @@ impl<'m> FruitsPP<'m> {
             .n_droplets
             .unwrap_or_else(|| attributes.n_droplets.saturating_sub(self.n_misses));
 
+        let max_combo = attributes.max_combo();
+
         let n_fruits = self.n_fruits.unwrap_or_else(|| {
-            attributes
-                .max_combo
+            max_combo
                 .saturating_sub(self.n_misses)
                 .saturating_sub(n_droplets)
         });
@@ -171,7 +195,7 @@ impl<'m> FruitsPP<'m> {
         acc /= 100.0;
 
         let n_tiny_droplets = self.n_tiny_droplets.unwrap_or_else(|| {
-            ((acc * (attributes.max_combo + max_tiny_droplets) as f32).round() as usize)
+            ((acc * (max_combo + max_tiny_droplets) as f64).round() as usize)
                 .saturating_sub(n_fruits)
                 .saturating_sub(n_droplets)
         });
@@ -186,11 +210,13 @@ impl<'m> FruitsPP<'m> {
         self
     }
 
-    fn assert_hitresults(&mut self, attributes: &DifficultyAttributes) {
+    fn assert_hitresults(self, attributes: FruitsDifficultyAttributes) -> FruitsPPInner {
+        let max_combo = attributes.max_combo();
+
         let correct_combo_hits = self
             .n_fruits
             .and_then(|f| self.n_droplets.map(|d| f + d + self.n_misses))
-            .filter(|h| *h == attributes.max_combo);
+            .filter(|h| *h == max_combo);
 
         let correct_fruits = self
             .n_fruits
@@ -216,8 +242,7 @@ impl<'m> FruitsPP<'m> {
             let mut n_tiny_droplets = self.n_tiny_droplets.unwrap_or(0);
             let n_tiny_droplet_misses = self.n_tiny_droplet_misses.unwrap_or(0);
 
-            let missing = attributes
-                .max_combo
+            let missing = max_combo
                 .saturating_sub(n_fruits)
                 .saturating_sub(n_droplets)
                 .saturating_sub(self.n_misses);
@@ -232,26 +257,57 @@ impl<'m> FruitsPP<'m> {
                 .saturating_sub(n_tiny_droplets)
                 .saturating_sub(n_tiny_droplet_misses);
 
-            self.n_fruits.replace(n_fruits);
-            self.n_droplets.replace(n_droplets);
-            self.n_tiny_droplets.replace(n_tiny_droplets);
-            self.n_tiny_droplet_misses.replace(n_tiny_droplet_misses);
+            return FruitsPPInner {
+                attributes,
+                mods: self.mods,
+                combo: self.combo,
+                n_fruits,
+                n_droplets,
+                n_tiny_droplets,
+                n_tiny_droplet_misses,
+                n_misses: self.n_misses,
+            };
+        }
+
+        FruitsPPInner {
+            attributes,
+            mods: self.mods,
+            combo: self.combo,
+            n_fruits: self.n_fruits.unwrap_or(0),
+            n_droplets: self.n_droplets.unwrap_or(0),
+            n_tiny_droplets: self.n_tiny_droplets.unwrap_or(0),
+            n_tiny_droplet_misses: self.n_tiny_droplet_misses.unwrap_or(0),
+            n_misses: self.n_misses,
         }
     }
 
-    /// Returns an object which contains the pp and [`DifficultyAttributes`](crate::fruits::DifficultyAttributes)
-    /// containing stars and other attributes.
-    pub fn calculate(mut self) -> PpResult {
-        let attributes = self.attributes.take().unwrap_or_else(|| {
-            stars(self.map, self.mods, self.passed_objects)
-                .attributes()
-                .unwrap()
-        });
+    /// Calculate all performance related values, including pp and stars.
+    pub fn calculate(mut self) -> FruitsPerformanceAttributes {
+        let attributes = self
+            .attributes
+            .take()
+            .unwrap_or_else(|| stars(self.map, self.mods, self.passed_objects));
 
-        // Make sure all objects are set
-        self.assert_hitresults(&attributes);
+        self.assert_hitresults(attributes).calculate()
+    }
+}
 
+struct FruitsPPInner {
+    attributes: FruitsDifficultyAttributes,
+    mods: u32,
+    combo: Option<usize>,
+    n_fruits: usize,
+    n_droplets: usize,
+    n_tiny_droplets: usize,
+    n_tiny_droplet_misses: usize,
+    n_misses: usize,
+}
+
+impl FruitsPPInner {
+    fn calculate(self) -> FruitsPerformanceAttributes {
+        let attributes = &self.attributes;
         let stars = attributes.stars;
+        let max_combo = attributes.max_combo();
 
         // Relying heavily on aim
         let mut pp = (5.0 * (stars / 0.0049).max(1.0) - 4.0).powi(2) / 100_000.0;
@@ -259,30 +315,29 @@ impl<'m> FruitsPP<'m> {
         let mut combo_hits = self.combo_hits();
 
         if combo_hits == 0 {
-            combo_hits = attributes.max_combo;
+            combo_hits = max_combo;
         }
 
         // Longer maps are worth more
         let len_bonus = 0.95
-            + 0.3 * (combo_hits as f32 / 2500.0).min(1.0)
-            + (combo_hits > 2500) as u8 as f32 * (combo_hits as f32 / 2500.0).log10() * 0.475;
+            + 0.3 * (combo_hits as f64 / 2500.0).min(1.0)
+            + (combo_hits > 2500) as u8 as f64 * (combo_hits as f64 / 2500.0).log10() * 0.475;
+
         pp *= len_bonus;
 
         // Penalize misses exponentially
-        pp *= 0.97_f32.powi(self.n_misses as i32);
+        pp *= 0.97_f64.powi(self.n_misses as i32);
 
         // Combo scaling
-        if let Some(combo) = self.combo.filter(|_| attributes.max_combo > 0) {
-            pp *= (combo as f32 / attributes.max_combo as f32)
-                .powf(0.8)
-                .min(1.0);
+        if let Some(combo) = self.combo.filter(|_| max_combo > 0) {
+            pp *= (combo as f64 / max_combo as f64).powf(0.8).min(1.0);
         }
 
         // AR scaling
         let ar = attributes.ar;
         let mut ar_factor = 1.0;
         if ar > 9.0 {
-            ar_factor += 0.1 * (ar - 9.0) + (ar > 10.0) as u8 as f32 * 0.1 * (ar - 10.0);
+            ar_factor += 0.1 * (ar - 9.0) + (ar > 10.0) as u8 as f64 * 0.1 * (ar - 10.0);
         } else if ar < 8.0 {
             ar_factor += 0.025 * (8.0 - ar);
         }
@@ -310,57 +365,64 @@ impl<'m> FruitsPP<'m> {
             pp *= 0.9;
         }
 
-        PpResult {
+        FruitsPerformanceAttributes {
+            difficulty: self.attributes,
             pp,
-            attributes: StarResult::Fruits(attributes),
         }
     }
 
     #[inline]
     fn combo_hits(&self) -> usize {
-        self.n_fruits.unwrap_or(0) + self.n_droplets.unwrap_or(0) + self.n_misses
+        self.n_fruits + self.n_droplets + self.n_misses
     }
 
     #[inline]
     fn successful_hits(&self) -> usize {
-        self.n_fruits.unwrap_or(0)
-            + self.n_droplets.unwrap_or(0)
-            + self.n_tiny_droplets.unwrap_or(0)
+        self.n_fruits + self.n_droplets + self.n_tiny_droplets
     }
 
     #[inline]
     fn total_hits(&self) -> usize {
-        self.successful_hits() + self.n_tiny_droplet_misses.unwrap_or(0) + self.n_misses
+        self.successful_hits() + self.n_tiny_droplet_misses + self.n_misses
     }
 
     #[inline]
-    fn acc(&self) -> f32 {
+    fn acc(&self) -> f64 {
         let total_hits = self.total_hits();
 
         if total_hits == 0 {
             1.0
         } else {
-            (self.successful_hits() as f32 / total_hits as f32)
+            (self.successful_hits() as f64 / total_hits as f64)
                 .max(0.0)
                 .min(1.0)
         }
     }
 }
 
+/// Abstract type to provide flexibility when passing difficulty attributes to a performance calculation.
 pub trait FruitsAttributeProvider {
-    fn attributes(self) -> Option<DifficultyAttributes>;
+    /// Provide the actual difficulty attributes.
+    fn attributes(self) -> Option<FruitsDifficultyAttributes>;
 }
 
-impl FruitsAttributeProvider for DifficultyAttributes {
+impl FruitsAttributeProvider for FruitsDifficultyAttributes {
     #[inline]
-    fn attributes(self) -> Option<DifficultyAttributes> {
+    fn attributes(self) -> Option<FruitsDifficultyAttributes> {
         Some(self)
     }
 }
 
-impl FruitsAttributeProvider for StarResult {
+impl FruitsAttributeProvider for FruitsPerformanceAttributes {
     #[inline]
-    fn attributes(self) -> Option<DifficultyAttributes> {
+    fn attributes(self) -> Option<FruitsDifficultyAttributes> {
+        Some(self.difficulty)
+    }
+}
+
+impl FruitsAttributeProvider for DifficultyAttributes {
+    #[inline]
+    fn attributes(self) -> Option<FruitsDifficultyAttributes> {
         #[allow(irrefutable_let_patterns)]
         if let Self::Fruits(attributes) = self {
             Some(attributes)
@@ -370,10 +432,15 @@ impl FruitsAttributeProvider for StarResult {
     }
 }
 
-impl FruitsAttributeProvider for PpResult {
+impl FruitsAttributeProvider for PerformanceAttributes {
     #[inline]
-    fn attributes(self) -> Option<DifficultyAttributes> {
-        self.attributes.attributes()
+    fn attributes(self) -> Option<FruitsDifficultyAttributes> {
+        #[allow(irrefutable_let_patterns)]
+        if let Self::Fruits(attributes) = self {
+            Some(attributes.difficulty)
+        } else {
+            None
+        }
     }
 }
 
@@ -382,12 +449,11 @@ mod test {
     use super::*;
     use crate::Beatmap;
 
-    fn attributes() -> DifficultyAttributes {
-        DifficultyAttributes {
+    fn attributes() -> FruitsDifficultyAttributes {
+        FruitsDifficultyAttributes {
             n_fruits: 1234,
             n_droplets: 567,
             n_tiny_droplets: 2345,
-            max_combo: 1234 + 567,
             ..Default::default()
         }
     }
@@ -410,7 +476,7 @@ mod test {
             + calculator.n_tiny_droplets.unwrap_or(0);
         let denominator =
             numerator + calculator.n_tiny_droplet_misses.unwrap_or(0) + calculator.n_misses;
-        let acc = 100.0 * numerator as f32 / denominator as f32;
+        let acc = 100.0 * numerator as f64 / denominator as f64;
 
         assert!(
             (target_acc - acc).abs() < 1.0,
@@ -450,7 +516,7 @@ mod test {
             + calculator.n_tiny_droplets.unwrap_or(0);
         let denominator =
             numerator + calculator.n_tiny_droplet_misses.unwrap_or(0) + calculator.n_misses;
-        let acc = 100.0 * numerator as f32 / denominator as f32;
+        let acc = 100.0 * numerator as f64 / denominator as f64;
 
         assert!(
             (target_acc - acc).abs() < 1.0,
@@ -472,42 +538,38 @@ mod test {
         let n_tiny_droplet_misses = 20;
         let n_misses = 2;
 
-        let mut calculator = FruitsPP::new(&map)
+        let calculator = FruitsPP::new(&map)
             .attributes(attributes.clone())
             .passed_objects(total_objects)
             .fruits(n_fruits)
             .droplets(n_droplets)
             .tiny_droplets(n_tiny_droplets)
             .tiny_droplet_misses(n_tiny_droplet_misses)
-            .misses(n_misses);
-
-        calculator.assert_hitresults(&attributes);
+            .misses(n_misses)
+            .assert_hitresults(attributes.clone());
 
         assert!(
-            (attributes.n_fruits as i32 - calculator.n_fruits.unwrap() as i32).abs()
-                <= n_misses as i32,
+            (attributes.n_fruits as i32 - calculator.n_fruits as i32).abs() <= n_misses as i32,
             "Expected: {} | Actual: {} [+/- {} misses]",
             attributes.n_fruits,
-            calculator.n_fruits.unwrap(),
+            calculator.n_fruits,
             n_misses
         );
 
         assert_eq!(
             attributes.n_droplets,
-            calculator.n_droplets.unwrap()
-                - (n_misses - (attributes.n_fruits - calculator.n_fruits.unwrap())),
+            calculator.n_droplets - (n_misses - (attributes.n_fruits - calculator.n_fruits)),
             "Expected: {} | Actual: {}",
             attributes.n_droplets,
-            calculator.n_droplets.unwrap()
-                - (n_misses - (attributes.n_fruits - calculator.n_fruits.unwrap())),
+            calculator.n_droplets - (n_misses - (attributes.n_fruits - calculator.n_fruits)),
         );
 
         assert_eq!(
             attributes.n_tiny_droplets,
-            calculator.n_tiny_droplets.unwrap() + calculator.n_tiny_droplet_misses.unwrap(),
+            calculator.n_tiny_droplets + calculator.n_tiny_droplet_misses,
             "Expected: {} | Actual: {}",
             attributes.n_tiny_droplets,
-            calculator.n_tiny_droplets.unwrap() + calculator.n_tiny_droplet_misses.unwrap(),
+            calculator.n_tiny_droplets + calculator.n_tiny_droplet_misses,
         );
     }
 }
