@@ -1,9 +1,3 @@
-#[cfg(any(
-    feature = "fruits",
-    all(feature = "osu", not(feature = "no_sliders_no_leniency"))
-))]
-use crate::math_util;
-
 mod attributes;
 mod control_point;
 mod error;
@@ -21,52 +15,63 @@ pub use pos2::Pos2;
 use sort::legacy_sort;
 
 use std::cmp::Ordering;
-use std::str::FromStr;
 
 #[cfg(not(any(feature = "async_std", feature = "async_tokio")))]
-use std::io::{BufRead, BufReader, Read};
+use std::{
+    fs::File,
+    io::{BufRead, BufReader, Read},
+};
 
 #[cfg(feature = "async_tokio")]
-use tokio::io::{AsyncBufReadExt, AsyncRead, BufReader};
+use tokio::{
+    fs::File,
+    io::{AsyncBufReadExt, AsyncRead, BufReader},
+};
+
+#[cfg(not(feature = "async_std"))]
+use std::path::Path;
 
 #[cfg(feature = "async_std")]
-use async_std::io::{prelude::BufReadExt, BufReader as AsyncBufReader, Read as AsyncRead};
+use async_std::{
+    fs::File,
+    io::{prelude::BufReadExt, BufReader as AsyncBufReader, Read as AsyncRead},
+    path::Path,
+};
 
-macro_rules! sort {
-    ($slice:expr) => {
-        $slice.sort_unstable_by(|p1, p2| p1.partial_cmp(&p2).unwrap_or(Ordering::Equal))
-    };
+#[cfg(feature = "sliders")]
+pub use osu_fruits::*;
 
-    (stable $slice:expr) => {
-        $slice.sort_by(|p1, p2| p1.partial_cmp(&p2).unwrap_or(Ordering::Equal))
-    };
+fn sort_unstable<T: PartialOrd>(slice: &mut [T]) {
+    slice.sort_unstable_by(|p1, p2| p1.partial_cmp(p2).unwrap_or(Ordering::Equal));
 }
 
-macro_rules! next_field {
-    ($opt:expr, $err:literal) => {
-        $opt.ok_or_else(|| ParseError::MissingField($err))?
-    };
+trait OptionExt<T> {
+    fn next_field(self, field: &'static str) -> Result<T, ParseError>;
 }
 
-macro_rules! validate_float {
-    ($x:expr) => {{
-        if $x.is_finite() {
-            $x
-        } else {
-            return Err(ParseError::InvalidFloatingPoint);
-        }
-    }};
+impl<T> OptionExt<T> for Option<T> {
+    fn next_field(self, field: &'static str) -> Result<T, ParseError> {
+        self.ok_or(ParseError::MissingField(field))
+    }
+}
+
+trait FloatExt: Sized {
+    fn validate(self) -> Result<Self, ParseError>;
+}
+
+impl FloatExt for f64 {
+    fn validate(self) -> Result<Self, ParseError> {
+        self.is_finite()
+            .then(|| self)
+            .ok_or(ParseError::InvalidDecimalNumber)
+    }
 }
 
 macro_rules! line_prepare {
     ($buf:ident) => {{
         let mut line = $buf.trim_end();
 
-        if line.is_empty()
-            || line.starts_with("//")
-            || line.starts_with(' ')
-            || line.starts_with('_')
-        {
+        if skip_line(line) {
             $buf.clear();
             continue;
         }
@@ -112,7 +117,7 @@ macro_rules! parse_general_body {
         let mut mode = None;
         let mut empty = true;
 
-        #[cfg(all(feature = "osu", feature = "all_included"))]
+        #[cfg(feature = "osu")]
         let mut stack_leniency = None;
 
         while read_line!($reader, $buf)? != 0 {
@@ -137,7 +142,7 @@ macro_rules! parse_general_body {
                 };
             }
 
-            #[cfg(all(feature = "osu", feature = "all_included"))]
+            #[cfg(feature = "osu")]
             if key == "StackLeniency" {
                 stack_leniency = Some(value.parse()?);
             }
@@ -167,7 +172,7 @@ macro_rules! parse_general_body {
             return Err(ParseError::UnincludedMode(GameMode::MNA));
         }
 
-        #[cfg(all(feature = "osu", feature = "all_included"))]
+        #[cfg(feature = "osu")]
         {
             $self.stack_leniency = stack_leniency.unwrap_or(0.7);
         }
@@ -177,10 +182,10 @@ macro_rules! parse_general_body {
 }
 
 macro_rules! parse_general {
-    ($reader:ident<$inner:ident>) => {
-        fn parse_general<R: $inner>(
+    () => {
+        fn parse_general<R: Read>(
             &mut self,
-            reader: &mut $reader<R>,
+            reader: &mut BufReader<R>,
             buf: &mut String,
             section: &mut Section,
         ) -> ParseResult<bool> {
@@ -236,22 +241,22 @@ macro_rules! parse_difficulty_body {
             $buf.clear();
         }
 
-        $self.od = next_field!(od, "od");
-        $self.cs = next_field!(cs, "cs");
-        $self.hp = next_field!(hp, "hp");
+        $self.od = od.next_field("od")?;
+        $self.cs = cs.next_field("cs")?;
+        $self.hp = hp.next_field("hp")?;
         $self.ar = ar.unwrap_or($self.od);
-        $self.sv = next_field!(sv, "sv");
-        $self.tick_rate = next_field!(tick_rate, "sv");
+        $self.slider_mult = sv.next_field("sv")?;
+        $self.tick_rate = tick_rate.next_field("tick rate")?;
 
         Ok(empty)
     }};
 }
 
 macro_rules! parse_difficulty {
-    ($reader:ident<$inner:ident>) => {
-        fn parse_difficulty<R: $inner>(
+    () => {
+        fn parse_difficulty<R: Read>(
             &mut self,
-            reader: &mut $reader<R>,
+            reader: &mut BufReader<R>,
             buf: &mut String,
             section: &mut Section,
         ) -> ParseResult<bool> {
@@ -275,6 +280,42 @@ macro_rules! parse_timingpoints_body {
     (short => $self:ident, $reader:ident, $buf:ident, $section:ident) => {{
         let mut empty = true;
 
+        // Only parse the first timing point to calculate the bpm
+        if read_line!($reader, $buf)? != 0 {
+            let line = {
+                let mut line = $buf.trim_end();
+
+                if skip_line(line) {
+                    $buf.clear();
+                    return Ok(empty);
+                }
+
+                if let Some(idx) = line.find("//") {
+                    line = &line[..idx];
+                }
+
+                line
+            };
+
+            if line.starts_with('[') && line.ends_with(']') {
+                *$section = Section::from_str(&line[1..line.len() - 1]);
+                empty = false;
+                $buf.clear();
+                return Ok(empty);
+            }
+
+            let beat_len = line
+                .split(',')
+                .nth(1)
+                .next_field("beat_len")?
+                .trim()
+                .parse()?;
+
+            $self.bpm = bpm(beat_len);
+
+            $buf.clear();
+        }
+
         while read_line!($reader, $buf)? != 0 {
             let line = line_prepare!($buf);
 
@@ -312,19 +353,19 @@ macro_rules! parse_timingpoints_body {
 
             let mut split = line.split(',');
 
-            let time = next_field!(split.next(), "timing point time")
+            let time = split
+                .next()
+                .next_field("timing point time")?
                 .trim()
-                .parse::<f32>()?;
-            validate_float!(time);
+                .parse::<f64>()?
+                .validate()?;
 
-            let beat_len = next_field!(split.next(), "beat len")
-                .trim()
-                .parse::<f32>()?;
+            let beat_len: f64 = split.next().next_field("beat len")?.trim().parse()?;
 
             if beat_len < 0.0 {
                 let point = DifficultyPoint {
                     time,
-                    speed_multiplier: -100.0 / beat_len,
+                    speed_multiplier: (-100.0 / beat_len).max(0.1).min(10.0),
                 };
 
                 $self.difficulty_points.push(point);
@@ -348,11 +389,11 @@ macro_rules! parse_timingpoints_body {
         }
 
         if unsorted_timings {
-            sort!($self.timing_points);
+            sort_unstable(&mut $self.timing_points);
         }
 
         if unsorted_difficulties {
-            sort!($self.difficulty_points);
+            sort_unstable(&mut $self.difficulty_points);
         }
 
         Ok(empty)
@@ -360,19 +401,19 @@ macro_rules! parse_timingpoints_body {
 }
 
 macro_rules! parse_timingpoints {
-    ($reader:ident<$inner:ident>) => {
-        fn parse_timingpoints<R: $inner>(
+    () => {
+        fn parse_timingpoints<R: Read>(
             &mut self,
-            reader: &mut $reader<R>,
+            reader: &mut BufReader<R>,
             buf: &mut String,
             section: &mut Section,
         ) -> ParseResult<bool> {
-            #[cfg(not(any(feature = "osu", feature = "fruits")))]
+            #[cfg(not(feature = "sliders"))]
             {
                 parse_timingpoints_body!(short => self, reader, buf, section)
             }
 
-            #[cfg(any(feature = "osu", feature = "fruits"))]
+            #[cfg(feature = "sliders")]
             parse_timingpoints_body!(self, reader, buf, section)
         }
     };
@@ -384,12 +425,12 @@ macro_rules! parse_timingpoints {
             buf: &mut String,
             section: &mut Section,
         ) -> ParseResult<bool> {
-            #[cfg(not(any(feature = "osu", feature = "fruits")))]
+            #[cfg(not(feature = "sliders"))]
             {
                 parse_timingpoints_body!(short => self, reader, buf, section)
             }
 
-            #[cfg(any(feature = "osu", feature = "fruits"))]
+            #[cfg(feature = "sliders")]
             parse_timingpoints_body!(self, reader, buf, section)
         }
     };
@@ -399,9 +440,19 @@ macro_rules! parse_hitobjects_body {
     ($self:ident, $reader:ident, $buf:ident, $section:ident) => {{
         let mut unsorted = false;
         let mut prev_time = 0.0;
-
         let mut empty = true;
 
+        // `point_split` will be of type `Vec<&str>
+        // with each element having its lifetime bound to `buf`.
+        // To circumvent this, `point_split_raw` will contain
+        // the actual `&str` elements transmuted into `usize`.
+        #[cfg(feature = "sliders")]
+        let mut point_split_raw: Vec<usize> = Vec::new();
+
+        // Buffer to re-use for all sliders
+        #[cfg(feature = "sliders")]
+        let mut vertices = Vec::new();
+
         while read_line!($reader, $buf)? != 0 {
             let line = line_prepare!($buf);
 
@@ -415,21 +466,22 @@ macro_rules! parse_hitobjects_body {
             let mut split = line.split(',');
 
             let pos = Pos2 {
-                x: next_field!(split.next(), "x position").parse()?,
-                y: next_field!(split.next(), "y position").parse()?,
+                x: split.next().next_field("x pos")?.parse()?,
+                y: split.next().next_field("y pos")?.parse()?,
             };
 
-            let time = next_field!(split.next(), "hitobject time")
+            let time = split
+                .next()
+                .next_field("hitobject time")?
                 .trim()
-                .parse::<f32>()?;
-
-            validate_float!(time);
+                .parse::<f64>()?
+                .validate()?;
 
             if !$self.hit_objects.is_empty() && time < prev_time {
                 unsorted = true;
             }
 
-            let kind: u8 = next_field!(split.next(), "hitobject kind").parse()?;
+            let kind: u8 = split.next().next_field("hitobject kind")?.parse()?;
             let sound = split.next().map(str::parse).transpose()?.unwrap_or(0);
 
             let kind = if kind & Self::CIRCLE_FLAG > 0 {
@@ -439,100 +491,104 @@ macro_rules! parse_hitobjects_body {
             } else if kind & Self::SLIDER_FLAG > 0 {
                 $self.n_sliders += 1;
 
-                #[cfg(any(
-                    feature = "fruits",
-                    all(feature = "osu", not(feature = "no_sliders_no_leniency"))
-                ))]
+                #[cfg(feature = "sliders")]
                 {
-                    let mut curve_points = Vec::with_capacity(4);
-                    curve_points.push(pos);
+                    let mut control_points = Vec::new();
 
-                    let mut curve_point_iter = next_field!(split.next(), "curve points").split('|');
+                    let control_point_iter = split.next().next_field("control points")?.split('|');
+                    let mut repeats: usize = split.next().next_field("repeats")?.parse()?;
 
-                    let mut path_type: PathType =
-                        next_field!(curve_point_iter.next(), "path kind").parse()?;
-
-                    for pos in curve_point_iter {
-                        let mut v = pos.split(':').map(str::parse);
-
-                        match (v.next(), v.next()) {
-                            (Some(Ok(x)), Some(Ok(y))) => curve_points.push(Pos2 { x, y }),
-                            _ => return Err(ParseError::InvalidCurvePoints),
-                        }
+                    if repeats > 9000 {
+                        return Err(ParseError::TooManyRepeats);
                     }
 
-                    match path_type {
-                        PathType::Linear if curve_points.len() % 2 == 0 => {
-                            // Assert that the points are of the form A|B|B|C|C|E
-                            if math_util::valid_linear(&curve_points) {
-                                for i in (2..curve_points.len() - 1).rev().step_by(2) {
-                                    curve_points.remove(i);
-                                }
-                            } else {
-                                path_type = PathType::Bezier;
-                            }
-                        }
-                        PathType::PerfectCurve if curve_points.len() == 3 => {
-                            if math_util::is_linear(curve_points[0], curve_points[1], curve_points[2]) {
-                                path_type = PathType::Linear;
-                            }
-                        },
-                        PathType::Catmull => {},
-                        _ => path_type = PathType::Bezier,
-                    };
+                    // * osu-stable treated the first span of the slider
+                    // * as a repeat, but no repeats are happening
+                    repeats = repeats.saturating_sub(1);
 
-                    // Reduce amount of curvepoints but keep the elements evenly spaced.
-                    // Necessary to handle maps like XNOR (2573164) which have
-                    // tens of thousands of curvepoints more efficiently.
-                    while curve_points.len() > CURVE_POINT_THRESHOLD {
-                        let last = curve_points[curve_points.len() - 1];
-                        let last_idx = (curve_points.len() - 1) / 2;
+                    let mut start_idx = 0;
+                    let mut end_idx = 0;
+                    let mut first = true;
 
-                        for i in 1..=last_idx {
-                            curve_points.swap(i, 2 * i);
+                    // SAFETY: `Vec<usize>` and `Vec<&str>` have the same size and layout.
+                    let point_split: &mut Vec<&str> =
+                        unsafe { std::mem::transmute(&mut point_split_raw) };
+
+                    point_split.clear();
+                    point_split.extend(control_point_iter);
+
+                    #[allow(clippy::blocks_in_if_conditions)]
+                    while {
+                        end_idx += 1;
+
+                        end_idx < point_split.len()
+                    } {
+                        // * Keep incrementing end_idx while it's not the start of a new segment
+                        // * (indicated by having a type descriptor of length 1).
+                        if point_split[end_idx].len() > 1 {
+                            continue;
                         }
 
-                        curve_points[last_idx] = last;
-                        curve_points.truncate(last_idx + 1);
+                        // * Multi-segmented sliders DON'T contain the end point as part of the
+                        // * current segment as it's assumed to be the start of the next segment.
+                        // * The start of the next segment is the index after the type descriptor.
+                        let end_point = point_split.get(end_idx + 1).copied();
+
+                        convert_points(
+                            &point_split[start_idx..end_idx],
+                            end_point,
+                            first,
+                            pos,
+                            &mut control_points,
+                            &mut vertices,
+                        )?;
+
+                        start_idx = end_idx;
+                        first = false;
                     }
 
-                    if curve_points.is_empty() {
+                    if end_idx > start_idx {
+                        convert_points(
+                            &point_split[start_idx..end_idx],
+                            None,
+                            first,
+                            pos,
+                            &mut control_points,
+                            &mut vertices,
+                        )?;
+                    }
+
+                    if control_points.is_empty() {
                         HitObjectKind::Circle
                     } else {
-                        let repeats = next_field!(split.next(), "repeats")
-                            .parse::<usize>()?
-                            .min(9000);
-
-                        let pixel_len = next_field!(split.next(), "pixel len")
-                            .parse::<f32>()?
+                        let pixel_len = split
+                            .next()
+                            .next_field("pixel len")?
+                            .parse::<f64>()?
                             .max(0.0)
                             .min(MAX_COORDINATE_VALUE);
 
                         HitObjectKind::Slider {
                             repeats,
                             pixel_len,
-                            curve_points,
-                            path_type,
+                            control_points,
                         }
                     }
                 }
 
-                #[cfg(not(any(
-                    feature = "fruits",
-                    all(feature = "osu", not(feature = "no_sliders_no_leniency"))
-                )))]
+                #[cfg(not(feature = "sliders"))]
                 {
-                    let repeats = next_field!(split.nth(1), "repeats").parse::<usize>()?;
-                    let len: f32 = next_field!(split.next(), "pixel len").parse()?;
+                    let span_count = split.nth(1).next_field("repeats")?.parse()?;
+                    let pixel_len = split.next().next_field("pixel len")?.parse()?;
 
                     HitObjectKind::Slider {
-                        repeats,
-                        pixel_len: len,
+                        span_count,
+                        pixel_len,
                     }
                 }
             } else if kind & Self::SPINNER_FLAG > 0 {
                 $self.n_spinners += 1;
-                let end_time = next_field!(split.next(), "spinner endtime").parse()?;
+                let end_time = split.next().next_field("spinner endtime")?.parse()?;
 
                 HitObjectKind::Spinner { end_time }
             } else if kind & Self::HOLD_FLAG > 0 {
@@ -540,7 +596,7 @@ macro_rules! parse_hitobjects_body {
                 let mut end = time;
 
                 if let Some(next) = split.next() {
-                    end = end.max(next_field!(next.split(':').next(), "hold endtime").parse()?);
+                    end = end.max(next.split(':').next().next_field("hold endtime")?.parse()?);
                 }
 
                 HitObjectKind::Hold { end_time: end }
@@ -563,12 +619,14 @@ macro_rules! parse_hitobjects_body {
         // won't be set yet so mania objects won't be sorted properly
         if $self.mode == GameMode::MNA {
             // First a _stable_ sort by time
-            sort!(stable $self.hit_objects);
+            $self
+                .hit_objects
+                .sort_by(|p1, p2| p1.partial_cmp(p2).unwrap_or(Ordering::Equal));
 
             // Then the legacy sort for correct position order
             legacy_sort(&mut $self.hit_objects);
         } else if unsorted {
-            sort!($self.hit_objects);
+            sort_unstable(&mut $self.hit_objects);
         }
 
         Ok(empty)
@@ -576,10 +634,10 @@ macro_rules! parse_hitobjects_body {
 }
 
 macro_rules! parse_hitobjects {
-    ($reader:ident<$inner:ident>) => {
-        fn parse_hitobjects<R: $inner>(
+    () => {
+        fn parse_hitobjects<R: Read>(
             &mut self,
-            reader: &mut $reader<R>,
+            reader: &mut BufReader<R>,
             buf: &mut String,
             section: &mut Section,
         ) -> ParseResult<bool> {
@@ -658,26 +716,63 @@ macro_rules! parse_body {
 }
 
 macro_rules! parse {
-    ($reader:ident<$inner:ident>) => {
-        pub fn parse<R: $inner>(input: R) -> ParseResult<Self> {
-            parse_body!($reader<$inner>: input)
+    () => {
+        /// Parse a beatmap from a `.osu` file.
+        ///
+        /// As argument you can give anything that implements [`std::io::Read`].
+        /// You'll likely want to pass (a reference of) a [`File`](std::fs::File)
+        /// or the file's content as a slice of bytes (`&[u8]`).
+        pub fn parse<R: Read>(input: R) -> ParseResult<Self> {
+            parse_body!(BufReader<Read>: input)
         }
     };
 
     (async $reader:ident<$inner:ident>) => {
+        /// Parse a beatmap from a `.osu` file.
+        ///
+        /// As argument you can give anything that implements `tokio::io::AsyncRead`
+        /// or `async_std::io::Read`, depending which feature you chose.
+        /// You'll likely want to pass a `File`
+        /// or the file's content as a slice of bytes (`&[u8]`).
         pub async fn parse<R: $inner + Unpin>(input: R) -> ParseResult<Self> {
             parse_body!($reader<$inner>: input)
         }
     };
 }
 
+macro_rules! from_path {
+    () => {
+        /// Pass the path to a `.osu` file.
+        ///
+        /// Useful when you don't want to create the [`File`](std::fs::File) manually.
+        /// If you have the file lying around already though (and plan on re-using it),
+        /// passing `&file` to [`parse`](Beatmap::parse) should be preferred.
+        pub fn from_path<P: AsRef<Path>>(path: P) -> ParseResult<Self> {
+            Self::parse(File::open(path)?)
+        }
+    };
+
+    (async $path:ident) => {
+        /// Pass the path to a `.osu` file.
+        ///
+        /// Useful when you don't want to create the file manually.
+        pub async fn from_path<P: AsRef<$path>>(path: P) -> ParseResult<Self> {
+            Self::parse(File::open(path).await?).await
+        }
+    };
+}
+
 /// The mode of a beatmap.
 #[derive(Copy, Clone, Debug, Hash, PartialEq, Eq)]
 #[allow(clippy::upper_case_acronyms)]
 pub enum GameMode {
+    /// osu!standard
     STD = 0,
+    /// osu!taiko
     TKO = 1,
+    /// osu!ctb
     CTB = 2,
+    /// osu!mania
     MNA = 3,
 }
 
@@ -692,45 +787,53 @@ impl Default for GameMode {
 /// for difficulty and pp calculation
 #[derive(Clone, Default, Debug)]
 pub struct Beatmap {
+    /// The game mode.
     pub mode: GameMode,
+    /// The version of the .osu file.
     pub version: u8,
 
+    /// The amount of circles.
     pub n_circles: u32,
+    /// The amount of sliders.
     pub n_sliders: u32,
+    /// The amount of spinners.
     pub n_spinners: u32,
 
+    /// The approach rate.
     pub ar: f32,
+    /// The overall difficulty.
     pub od: f32,
+    /// The circle size.
     pub cs: f32,
+    /// The health drain rate.
     pub hp: f32,
-    pub sv: f32,
-    pub tick_rate: f32,
+    /// Base slider velocity in pixels per beat
+    pub slider_mult: f64,
+    /// Amount of slider ticks per beat.
+    pub tick_rate: f64,
+    /// All hitobjects of the beatmap.
     pub hit_objects: Vec<HitObject>,
 
-    #[cfg(any(feature = "osu", feature = "fruits"))]
+    #[cfg(not(feature = "sliders"))]
+    /// Beats per minute
+    pub bpm: f64,
+
+    #[cfg(feature = "sliders")]
+    /// Timing points that indicate a new timing section.
     pub timing_points: Vec<TimingPoint>,
 
-    #[cfg(any(feature = "osu", feature = "fruits"))]
+    #[cfg(feature = "sliders")]
+    /// Timing point for the current timing section.
     pub difficulty_points: Vec<DifficultyPoint>,
 
-    #[cfg(all(feature = "osu", feature = "all_included"))]
+    #[cfg(feature = "osu")]
+    /// The stack leniency that is used to calculate
+    /// the stack offset for stacked positions.
     pub stack_leniency: f32,
 }
 
 pub(crate) const OSU_FILE_HEADER: &str = "osu file format v";
 
-#[cfg(any(
-    feature = "fruits",
-    all(feature = "osu", not(feature = "no_sliders_no_leniency"))
-))]
-const CURVE_POINT_THRESHOLD: usize = 256;
-
-#[cfg(any(
-    feature = "fruits",
-    all(feature = "osu", not(feature = "no_sliders_no_leniency"))
-))]
-const MAX_COORDINATE_VALUE: f32 = 131_072.0;
-
 impl Beatmap {
     const CIRCLE_FLAG: u8 = 1 << 0;
     const SLIDER_FLAG: u8 = 1 << 1;
@@ -739,19 +842,189 @@ impl Beatmap {
     // const COMBO_OFFSET_FLAG: u8 = (1 << 4) | (1 << 5) | (1 << 6);
     const HOLD_FLAG: u8 = 1 << 7;
 
+    /// Extract a beatmap's attributes into their own type.
     #[inline]
     pub fn attributes(&self) -> BeatmapAttributes {
         BeatmapAttributes::new(self.ar, self.od, self.cs, self.hp)
     }
+
+    /// The beats per minute of the map.
+    #[cfg(feature = "sliders")]
+    #[inline]
+    pub fn bpm(&self) -> f64 {
+        match self.timing_points.first() {
+            Some(point) => bpm(point.beat_len),
+            None => 0.0,
+        }
+    }
+
+    /// The beats per minute of the map.
+    #[cfg(not(feature = "sliders"))]
+    #[inline]
+    pub fn bpm(&self) -> f64 {
+        self.bpm
+    }
+}
+
+#[cfg(feature = "sliders")]
+mod osu_fruits {
+    use crate::ParseError;
+
+    use super::Pos2;
+
+    pub(super) const MAX_COORDINATE_VALUE: f64 = 131_072.0;
+
+    pub(super) fn convert_points(
+        points: &[&str],
+        end_point: Option<&str>,
+        first: bool,
+        offset: Pos2,
+        curve_points: &mut Vec<PathControlPoint>,
+        vertices: &mut Vec<PathControlPoint>,
+    ) -> Result<(), ParseError> {
+        let mut path_kind = PathType::from_str(points[0]);
+
+        let read_offset = first as usize;
+        let readable_points = points.len() - 1;
+        let end_point_len = end_point.is_some() as usize;
+
+        vertices.clear();
+        vertices.reserve(read_offset + readable_points + end_point_len);
+
+        // * Fill any non-read points.
+        vertices.extend((0..read_offset).map(|_| PathControlPoint::default()));
+
+        // * Parse into control points.
+        for &point in points.iter().skip(1) {
+            vertices.push(read_point(point, offset)?);
+        }
+
+        // * If an endpoint is given, add it to the end.
+        if let Some(end_point) = end_point {
+            vertices.push(read_point(end_point, offset)?);
+        }
+
+        // * Edge-case rules (to match stable).
+        if path_kind == PathType::PerfectCurve {
+            if let [a, b, c] = &vertices[..] {
+                if is_linear(a.pos, b.pos, c.pos) {
+                    // * osu-stable special-cased colinear perfect curves to a linear path
+                    path_kind = PathType::Linear;
+                }
+            } else {
+                path_kind = PathType::Bezier;
+            }
+        }
+
+        // * The first control point must have a definite type.
+        vertices[0].kind = Some(path_kind);
+
+        // * A path can have multiple implicit segments of the same type if
+        // * there are two sequential control points with the same position.
+        // * To handle such cases, this code may return multiple path segments
+        // * with the final control point in each segment having a non-null type.
+        // * For the point string X|1:1|2:2|2:2|3:3, this code returns the segments:
+        // * X: { (1,1), (2, 2) }
+        // * X: { (3, 3) }
+        // * Note: (2, 2) is not returned in the second segments, as it is implicit in the path.
+        let mut start_idx = 0;
+        let mut end_idx = 0;
+
+        #[allow(clippy::blocks_in_if_conditions)]
+        while {
+            end_idx += 1;
+
+            end_idx < vertices.len() - end_point_len
+        } {
+            // * Keep incrementing while an implicit segment doesn't need to be started
+            if vertices[end_idx].pos != vertices[end_idx - 1].pos {
+                continue;
+            }
+
+            // * The last control point of each segment is not
+            // * allowed to start a new implicit segment.
+            if end_idx == vertices.len() - end_point_len - 1 {
+                continue;
+            }
+
+            // * Force a type on the last point, and return
+            // * the current control point set as a segment.
+            vertices[end_idx - 1].kind = Some(path_kind);
+            curve_points.extend(&vertices[start_idx..end_idx]);
+
+            // * Skip the current control point - as it's the same as the one that's just been returned.
+            start_idx = end_idx + 1;
+        }
+
+        if end_idx > start_idx {
+            curve_points.extend(&vertices[start_idx..end_idx]);
+        }
+
+        Ok(())
+    }
+
+    pub(super) fn read_point(value: &str, start_pos: Pos2) -> Result<PathControlPoint, ParseError> {
+        let mut v = value.split(':').map(str::parse);
+
+        match (v.next(), v.next()) {
+            (Some(Ok(x)), Some(Ok(y))) => Ok(PathControlPoint::from(Pos2 { x, y } - start_pos)),
+            _ => Err(ParseError::InvalidCurvePoints),
+        }
+    }
+
+    fn is_linear(p0: Pos2, p1: Pos2, p2: Pos2) -> bool {
+        ((p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)).abs() <= f32::EPSILON
+    }
+
+    /// Control point for slider curve calculation
+    #[derive(Copy, Clone, Debug, Default, PartialEq)]
+    pub struct PathControlPoint {
+        /// Control point position.
+        pub pos: Pos2,
+        /// Path type of the control point.
+        /// Only present for the first element of each segment.
+        pub kind: Option<PathType>,
+    }
+
+    impl From<Pos2> for PathControlPoint {
+        #[inline]
+        fn from(pos: Pos2) -> Self {
+            Self { pos, kind: None }
+        }
+    }
+
+    /// The type of curve of a slider.
+    #[allow(missing_docs)]
+    #[derive(Copy, Clone, Debug, Eq, PartialEq)]
+    pub enum PathType {
+        Catmull = 0,
+        Bezier = 1,
+        Linear = 2,
+        PerfectCurve = 3,
+    }
+
+    impl PathType {
+        #[inline]
+        fn from_str(s: &str) -> Self {
+            match s {
+                "L" => Self::Linear,
+                "B" => Self::Bezier,
+                "P" => Self::PerfectCurve,
+                _ => Self::Catmull,
+            }
+        }
+    }
 }
 
 #[cfg(not(any(feature = "async_std", feature = "async_tokio")))]
 impl Beatmap {
-    parse!(BufReader<Read>);
-    parse_general!(BufReader<Read>);
-    parse_difficulty!(BufReader<Read>);
-    parse_timingpoints!(BufReader<Read>);
-    parse_hitobjects!(BufReader<Read>);
+    parse!();
+    parse_general!();
+    parse_difficulty!();
+    parse_timingpoints!();
+    parse_hitobjects!();
+
+    from_path!();
 }
 
 #[cfg(feature = "async_tokio")]
@@ -761,6 +1034,8 @@ impl Beatmap {
     parse_difficulty!(async BufReader<AsyncRead>);
     parse_timingpoints!(async BufReader<AsyncRead>);
     parse_hitobjects!(async BufReader<AsyncRead>);
+
+    from_path!(async Path);
 }
 
 #[cfg(feature = "async_std")]
@@ -770,6 +1045,16 @@ impl Beatmap {
     parse_difficulty!(async AsyncBufReader<AsyncRead>);
     parse_timingpoints!(async AsyncBufReader<AsyncRead>);
     parse_hitobjects!(async AsyncBufReader<AsyncRead>);
+
+    from_path!(async Path);
+}
+
+fn bpm(beat_len: f64) -> f64 {
+    beat_len.recip() * 1000.0 * 60.0
+}
+
+fn skip_line(line: &str) -> bool {
+    line.is_empty() || line.starts_with("//") || line.starts_with(' ') || line.starts_with('_')
 }
 
 #[inline]
@@ -779,30 +1064,6 @@ fn split_colon(line: &str) -> Option<(&str, &str)> {
     Some((split.next()?, split.next()?.trim()))
 }
 
-/// The type of curve of a slider.
-#[derive(Copy, Clone, Debug, Eq, PartialEq)]
-pub enum PathType {
-    Catmull = 0,
-    Bezier = 1,
-    Linear = 2,
-    PerfectCurve = 3,
-}
-
-impl FromStr for PathType {
-    type Err = ParseError;
-
-    #[inline]
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        match s {
-            "L" => Ok(Self::Linear),
-            "C" => Ok(Self::Catmull),
-            "B" => Ok(Self::Bezier),
-            "P" => Ok(Self::PerfectCurve),
-            _ => Err(ParseError::InvalidPathType),
-        }
-    }
-}
-
 #[derive(Copy, Clone, Debug)]
 enum Section {
     None,
@@ -832,83 +1093,80 @@ mod tests {
     #[cfg(not(any(feature = "async_std", feature = "async_tokio")))]
     #[test]
     fn parsing_sync() {
-        use std::fs::File;
+        for map_id in map_ids() {
+            println!("map_id: {}", map_id);
 
-        let map_id = map_id();
-        println!("map_id: {}", map_id);
+            let map = match Beatmap::from_path(format!("./maps/{}.osu", map_id)) {
+                Ok(map) => map,
+                Err(why) => panic!("Error while parsing map: {}", why),
+            };
 
-        let file = match File::open(format!("./maps/{}.osu", map_id)) {
-            Ok(file) => file,
-            Err(why) => panic!("Could not read file: {}", why),
-        };
-
-        let map = match Beatmap::parse(file) {
-            Ok(map) => map,
-            Err(why) => panic!("Error while parsing map: {}", why),
-        };
-
-        print_info(map)
+            print_info(map);
+            println!("---");
+        }
     }
 
     #[cfg(feature = "async_tokio")]
     #[test]
     fn parsing_async_tokio() {
-        use tokio::{fs::File, runtime::Builder};
+        use tokio::runtime::Builder;
 
         Builder::new_current_thread()
             .build()
             .expect("could not start runtime")
             .block_on(async {
-                let map_id = map_id();
-                println!("map_id: {}", map_id);
+                for map_id in map_ids() {
+                    println!("map_id: {}", map_id);
 
-                let file = match File::open(format!("./maps/{}.osu", map_id)).await {
-                    Ok(file) => file,
-                    Err(why) => panic!("Could not read file: {}", why),
-                };
+                    let map = match Beatmap::from_path(format!("./maps/{}.osu", map_id)).await {
+                        Ok(map) => map,
+                        Err(why) => panic!("Error while parsing map: {}", why),
+                    };
 
-                let map = match Beatmap::parse(file).await {
-                    Ok(map) => map,
-                    Err(why) => panic!("Error while parsing map: {}", why),
-                };
-
-                print_info(map)
+                    print_info(map);
+                    println!("---");
+                }
             });
     }
 
     #[cfg(feature = "async_std")]
     #[test]
     fn parsing_async_std() {
-        use async_std::fs::File;
-
         async_std::task::block_on(async {
-            let map_id = map_id();
-            println!("map_id: {}", map_id);
+            for map_id in map_ids() {
+                println!("map_id: {}", map_id);
 
-            let file = match File::open(format!("./maps/{}.osu", map_id)).await {
-                Ok(file) => file,
-                Err(why) => panic!("Could not read file: {}", why),
-            };
+                let map = match Beatmap::from_path(format!("./maps/{}.osu", map_id)).await {
+                    Ok(map) => map,
+                    Err(why) => panic!("Error while parsing map: {}", why),
+                };
 
-            let map = match Beatmap::parse(file).await {
-                Ok(map) => map,
-                Err(why) => panic!("Error while parsing map: {}", why),
-            };
-
-            print_info(map)
+                print_info(map);
+                println!("---");
+            }
         });
     }
 
-    fn map_id() -> i32 {
+    fn map_ids() -> Vec<i32> {
+        let mut map_ids = Vec::new();
+
         if cfg!(feature = "osu") {
-            797130
-        } else if cfg!(feature = "mania") {
-            1355822
-        } else if cfg!(feature = "fruits") {
-            1977380
-        } else {
-            110219
+            map_ids.push(2785319);
         }
+
+        if cfg!(feature = "mania") {
+            map_ids.push(1974394);
+        }
+
+        if cfg!(feature = "fruits") {
+            map_ids.push(2118524);
+        }
+
+        if cfg!(feature = "taiko") {
+            map_ids.push(1028484);
+        }
+
+        map_ids
     }
 
     fn print_info(map: Beatmap) {
@@ -920,13 +1178,13 @@ mod tests {
         println!("od: {}", map.od);
         println!("cs: {}", map.cs);
         println!("hp: {}", map.hp);
-        println!("sv: {}", map.sv);
+        println!("slider_mult: {}", map.slider_mult);
         println!("tick_rate: {}", map.tick_rate);
         println!("hit_objects: {}", map.hit_objects.len());
 
-        #[cfg(any(feature = "osu", feature = "fruits"))]
+        #[cfg(feature = "sliders")]
         {
-            #[cfg(feature = "all_included")]
+            #[cfg(feature = "osu")]
             println!("stack_leniency: {}", map.stack_leniency);
 
             println!("timing_points: {}", map.timing_points.len());
